@@ -1,33 +1,37 @@
 package Test::LWP::UserAgent;
 {
-  $Test::LWP::UserAgent::VERSION = '0.007';
+  $Test::LWP::UserAgent::VERSION = '0.008';
 }
-# git description: v0.006-TRIAL-12-gfbbc912
+# git description: v0.007-13-g3da0314
 
+# ABSTRACT: a LWP::UserAgent suitable for simulating and testing network calls
 
 use strict;
 use warnings;
 
 use parent 'LWP::UserAgent';
-use Scalar::Util qw(blessed reftype);
+use Scalar::Util 'blessed';
 use Storable 'freeze';
 use HTTP::Request;
 use HTTP::Response;
 use URI;
 use HTTP::Date;
+use HTTP::Status qw(:constants status_message);
+use Try::Tiny;
 
-my $last_http_request_sent;
-my $last_http_response_received;
 my @response_map;
+my $network_fallback;
+my $last_useragent;
 
 sub new
 {
-    my $class = shift;
+    my ($class, %options) = @_;
 
-    my $self = $class->SUPER::new(@_);
+    my $self = $class->SUPER::new(%options);
     $self->{__last_http_request_sent} = undef;
     $self->{__last_http_response_received} = undef;
     $self->{__response_map} = [];
+    $self->{__network_fallback} = $options{network_fallback};
 
     # strips default User-Agent header added by LWP::UserAgent, to make it
     # easier to define literal HTTP::Requests to match against
@@ -40,25 +44,27 @@ sub map_response
 {
     my ($self, $request_description, $response) = @_;
 
-    if (not defined $response and not reftype $request_description and blessed $self)
+    if (not defined $response and blessed $self)
     {
         # mask a global domain mapping
-        my @indexes = grep {
-            not reftype $_->[0] and $_->[0] eq $request_description
-        } @{$self->{__response_map}};
+        my $matched;
+        foreach my $mapping (@{$self->{__response_map}})
+        {
+            if ($mapping->[0] eq $request_description)
+            {
+                $matched = 1;
+                undef $mapping->[1];
+            }
+        }
+
+        push @{$self->{__response_map}}, [ $request_description, undef ]
+            if not $matched;
 
-        if (@indexes)
-        {
-            undef @{$self->{__response_map}}[@indexes];
-        }
-        else
-        {
-            push @{$self->{__response_map}}, [ $request_description, undef ];
-        }
         return;
     }
 
-    warn "map_response: response is not an HTTP::Response, it's a " . blessed($response)
+    warn "map_response: response is not an HTTP::Response, it's a ",
+            (blessed($response) || 'non-object')
         unless eval { \&$response } or eval { $response->isa('HTTP::Response') };
 
     if (blessed $self)
@@ -71,6 +77,22 @@ sub map_response
     }
 }
 
+sub map_network_response
+{
+    my ($self, $request_description) = @_;
+
+    if (blessed $self)
+    {
+        push @{$self->{__response_map}},
+            [ $request_description, sub { $self->SUPER::send_request($_[0]) } ];
+    }
+    else
+    {
+        push @response_map,
+            [ $request_description, sub { LWP::UserAgent->new->send_request($_[0]) } ];
+    }
+}
+
 sub unmap_all
 {
     my ($self, $instance_only) = @_;
@@ -94,7 +116,7 @@ sub register_psgi
 
     return $self->map_response($domain, undef) if not defined $app;
 
-    warn "register_psgi: app is not a coderef, it's a " . ref($app)
+    warn "register_psgi: app is not a coderef, it's a ", ref($app)
         unless eval { \&$app };
 
     warn "register_psgi: did you forget to load HTTP::Message::PSGI?"
@@ -128,7 +150,9 @@ sub last_http_request_sent
     my $self = shift;
     return blessed($self)
         ? $self->{__last_http_request_sent}
-        : $last_http_request_sent;
+        : $last_useragent
+        ? $last_useragent->last_http_request_sent
+        : undef;
 }
 
 sub last_http_response_received
@@ -136,7 +160,29 @@ sub last_http_response_received
     my $self = shift;
     return blessed($self)
         ? $self->{__last_http_response_received}
-        : $last_http_response_received;
+        : $last_useragent
+        ? $last_useragent->last_http_response_received
+        : undef;
+}
+
+sub last_useragent
+{
+    return $last_useragent;
+}
+
+sub network_fallback
+{
+    my ($self, $value) = @_;
+
+    if (@_ == 1)
+    {
+        return blessed $self
+            ? $self->{__network_fallback}
+            : $network_fallback;
+    }
+
+    return $self->{__network_fallback} = $value if blessed $self;
+    $network_fallback = $value;
 }
 
 sub __is_regexp($);
@@ -161,48 +207,82 @@ sub send_request
             $matched_response = $response, last
                 if freeze($request) eq freeze($request_desc);
         }
-        elsif (not reftype $request_desc)
+        elsif (__is_regexp $request_desc)
         {
+            $matched_response = $response, last
+                if $uri =~ $request_desc;
+        }
+        else
+        {
+            $matched_response = $response, last
+                if eval { $request_desc->($request) };
+
             $uri = URI->new($uri) if not eval { $uri->isa('URI') };
             $matched_response = $response, last
                 if $uri->host eq $request_desc;
         }
-        elsif (eval { \&$request_desc })
-        {
-            $matched_response = $response, last
-                if $request_desc->($request);
-        }
-        elsif (__is_regexp $request_desc)
-        {
-            $matched_response = $response, last
-                if $request->uri =~ $request_desc;
-        }
-        else
-        {
-            warn 'unknown request type found in ' . blessed($self) . ' mapping!';
-        }
     }
 
-    $last_http_request_sent = $self->{__last_http_request_sent} = $request;
+    $last_useragent = $self;
+    $self->{__last_http_request_sent} = $request;
 
-    my $response = defined $matched_response ? $matched_response : HTTP::Response->new(404);
+    if (not defined $matched_response and
+        ($self->{__network_fallback} or $network_fallback))
+    {
+        my $response = $self->SUPER::send_request($request);
+        $self->{__last_http_response_received} = $response;
+        return $response;
+    }
+
+    my $response = defined $matched_response
+        ? $matched_response
+        : HTTP::Response->new(404);
 
     if (eval { \&$response })
     {
-        $response = $response->($request);
-
-        warn "response from coderef is not a HTTP::Response, it's a ", blessed($response)
-            unless eval { $response->isa('HTTP::Response') };
+        # emulates handling in LWP::UserAgent::send_request
+        if ($self->use_eval)
+        {
+            $response = try { $response->($request) }
+            catch {
+                my $exception = $_;
+                if (eval { $exception->isa('HTTP::Response') })
+                {
+                    $response = $exception;
+                }
+                else
+                {
+                    my $full = $exception;
+                    (my $status = $exception) =~ s/\n.*//s;
+                    $status =~ s/ at .* line \d+.*//s;  # remove file/line number
+                    my $code = ($status =~ s/^(\d\d\d)\s+//) ? $1 : HTTP_INTERNAL_SERVER_ERROR;
+                    $response = LWP::UserAgent::_new_response($request, $code, $status, $full);
+                }
+            }
+        }
+        else
+        {
+            $response = $response->($request);
+        }
     }
 
-    $last_http_response_received = $self->{__last_http_response_received} = $response;
+    if (not eval { $response->isa('HTTP::Response') })
+    {
+        warn "response from coderef is not a HTTP::Response, it's a ",
+            (blessed($response) || 'non-object');
+        $response = LWP::UserAgent::_new_response($request, HTTP_INTERNAL_SERVER_ERROR, status_message(HTTP_INTERNAL_SERVER_ERROR));
+    }
+    else
+    {
+        $response->request($request);  # record request for reference
+        $response->header("Client-Date" => HTTP::Date::time2str(time));
+    }
 
-    # bookkeeping that the real LWP::UserAgent does
-    $response->request($request);  # record request for reference
-    $response->header("Client-Date" => HTTP::Date::time2str(time));
     $self->run_handlers("response_done", $response);
     $self->progress("end", $response);
 
+    $self->{__last_http_response_received} = $response;
+
     return $response;
 }
 
@@ -212,7 +292,7 @@ sub __is_regexp($)
 }
 
 1;
-__END__
+
 
 =pod
 
@@ -222,7 +302,7 @@ Test::LWP::UserAgent - a LWP::UserAgent suitable for simulating and testing netw
 
 =head1 VERSION
 
-version 0.007
+version 0.008
 
 =head1 SYNOPSIS
 
@@ -280,20 +360,78 @@ And then:
 
     # <now test that your code responded to the 200 response properly...>
 
+This feature is useful for testing your PSGI apps (you may or may not find
+using L<Plack::Test> easier), or for simulating a server so as to test your
+client code.
+
+OR, you can route some or all requests through the network as normal, but
+still gain the hooks provided by this class to test what was sent and
+received:
+
+    my $useragent = Test::LWP::UserAgent->new(network_fallback => 1);
+
+or:
+
+    $useragent->map_network_response(qr/real.network.host/);
+
+    # ... generate a request...
+
+    # and then in your tests:
+    is(
+        $useragent->last_useragent->timeout,
+        180,
+        'timeout was overridden properly',
+    );
+    is(
+        $useragent->last_http_request_sent->uri,
+        'uri my code should have constructed',
+    );
+    is(
+        $useragent->last_http_response_received->code,
+        200,
+        'I should have gotten an OK response',
+    );
+
 One common mechanism to swap out the useragent implementation is via a
 lazily-built Moose attribute; if no override is provided at construction time,
 default to C<< LWP::UserAgent->new(%options) >>.
 
 =head1 METHODS
 
-All methods may be called on a specific object instance, or as a class method.
+=over
+
+=item * C<new>
+
+Accepts all options as in L<LWP::UserAgent>, including C<use_eval>, an
+undocumented boolean which is enabled by default. When set, sending the HTTP
+request is wrapped in an C<< eval {} >>, allowing all exceptions to be caught
+and an appropriate error response (usually HTTP 500) to be returned. You may
+want to unset this if you really want to test extraordinary errors within your
+networking code.  Normally, you should leave it alone, as L<LWP::UserAgent> and
+this module are capable of handling normal errors.
+
+Plus, this option is added:
+
+=over
+
+=item * C<< network_fallback => <boolean> >>
+
+If true, requests passing through this object that do not match a
+previously-configured mapping or registration will be directed to the network.
+(To only divert I<matched> requests rather than unmatched requests, use
+C<map_network_response>, see below.)
+
+This option is also available as a read/write accessor via
+C<< $useragent->network_fallback(<value?>) >>.
+
+=back
+
+All other methods may be called on a specific object instance, or as a class method.
 If called as on a blessed object, the action performed or data returned is
 limited to just that object; if called as a class method, the action or data is
 global.
 
-=over
-
-=item C<map_response($request_description, $http_response)>
+=item * C<map_response($request_description, $http_response)>
 
 With this method, you set up what L<HTTP::Response> should be returned for each
 request received.
@@ -302,7 +440,7 @@ The request match specification can be described in multiple ways:
 
 =over
 
-=item string
+=item * string
 
 The string is matched identically against the C<host> field of the L<URI> in the request.
 
@@ -310,7 +448,7 @@ Example:
 
     $test_ua->map_response('example.com', HTTP::Response->new(500));
 
-=item regexp
+=item * regexp
 
 The regexp is matched against the URI in the request.
 
@@ -319,7 +457,7 @@ Example:
     $test_ua->map_response(qr{foo/bar}, HTTP::Response->new(200));
     $test_ua->map_response(qr{baz/quux}, HTTP::Response->new(500));
 
-=item code
+=item * code
 
 An arbitrary coderef is passed a single argument, the L<HTTP::Request>, and
 returns a boolean indicating if there is a match.
@@ -331,7 +469,7 @@ returns a boolean indicating if there is a match.
         HTTP::Response->new(200),
     );
 
-=item L<HTTP::Request> object
+=item * L<HTTP::Request> object
 
 The L<HTTP::Request> object is matched identically (including all query
 parameters, headers etc) against the provided object.
@@ -355,7 +493,18 @@ Instance mappings take priority over global (class method) mappings - if no
 matches are found from mappings added to the instance, the global mappings are
 then examined. After no matches have been found, a 404 response is returned.
 
-=item C<unmap_all(instance_only?)>
+=item * C<map_network_response($request_description)>
+
+Same as C<map_response> above, only requests that match this description will
+not use a response that you specify, but instead uses a real L<LWP::UserAgent>
+to dispatch your request to the network.
+
+If called on an instance, all options passed to the constructor (e.g. timeout)
+are used for making the real network call. If called as a class method, a
+pristine L<LWP::UserAgent> object with no customized options will be used
+instead.
+
+=item * C<unmap_all(instance_only?)>
 
 When called as a class method, removes all mappings set up globally (across all
 objects). Mappings set up on an individual object will still remain.
@@ -365,7 +514,7 @@ this instance, unless a true value is passed as an argument, in which only
 mappings local to the object will be removed. (Any true value will do, so you
 can pass a meaningful string.)
 
-=item C<register_psgi($domain, $app)>
+=item * C<register_psgi($domain, $app)>
 
 Register a particular L<PSGI> app (code reference) to be used when requests
 for a domain are received (matches are made exactly against
@@ -383,7 +532,7 @@ calling C<< $test_ua->register_psgi($domain, $app) >> is equivalent to:
         sub { HTTP::Response->from_psgi($app->($_[0]->to_psgi)) },
     );
 
-=item C<unregister_psgi($domain, instance_only?)>
+=item * C<unregister_psgi($domain, instance_only?)>
 
 When called as a class method, removes a domain->PSGI app entry that had been
 registered globally.  Some mappings set up on an individual object may still
@@ -399,32 +548,59 @@ then add C<undef> as a mapping on your instance:
 
     $useragent->map_response($domain, undef);
 
-=item C<last_http_request_sent>
+=item * C<last_http_request_sent>
 
 The last L<HTTP::Request> object that this object (if called on an object) or
 module (if called as a class method) processed, whether or not it matched a
 mapping you set up earlier.
 
-=item C<last_http_response_received>
+=item * C<last_http_response_received>
 
 The last L<HTTP::Response> object that this module returned, as a result of a
 mapping you set up earlier with C<map_response>. You shouldn't normally need to
 use this, as you know what you responded with - you should instead be testing
 how your code reacted to receiving this response.
 
-=item C<send_request($request)>
+=item * C<last_useragent>
+
+The last Test::LWP::UserAgent object that was used to send a request.
+Obviously this only provides new information if called as a class method; you
+can use this if you don't have direct control over the useragent itself, to
+get the object that was used, to verify options such as the network timeout.
+
+=item * C<network_fallback>
+
+Getter/setter method for the network_fallback preference that will be used on
+this object (if called as an instance method), or globally, if called as a
+class method.  Note that the actual behaviour used on an object is the ORed
+value of the instance setting and the global setting.
+
+=item * C<send_request($request)>
 
 This is the only method from L<LWP::UserAgent> that has been overridden, which
 processes the L<HTTP::Request>, sends to the network, then creates the
 L<HTTP::Response> object from the reply received. Here, we loop through your
 local and global domain registrations, and local and global mappings (in this
 order) and returns the first match found; otherwise, a simple 404 response is
-returned.
+returned (unless C<network_fallback> was specified as a constructor option,
+in which case unmatched requests will be delivered to the network.)
 
 =back
 
 All other methods from L<LWP::UserAgent> are available unchanged.
 
+=head1 Use with SOAP requests
+
+To use this module when communicating with a SOAP server (either a real one,
+with live network requests, see above ... link here ..., or with one simulated
+with mapped responses), simply do this:
+
+    use SOAP::Lite;
+    use SOAP::Transport::HTTP;
+    $SOAP::Transport::HTTP::Client::USERAGENT_CLASS = 'Test::LWP::UserAgent';
+
+See also L<SOAP::Transport/CHANGING THE DEFAULT USERAGENT CLASS>.
+
 =head1 MOTIVATION
 
 Most mock libraries on the CPAN use L<Test::MockObject>, which is widely considered
@@ -442,18 +618,6 @@ throughout. It provides a method (C<last_http_request_sent>) to access the last
 L<HTTP::Request>, for testing things like the URI and headers that your code
 sent to L<LWP::UserAgent>.
 
-=head1 TODO (possibly)
-
-=over
-
-=item Option to locally or globally override useragent implementations via
-symbol table swap
-
-=item Ability to route certain requests through the real network, to gain the
-benefits of C<last_http_request_sent> and C<last_http_response_received>
-
-=back
-
 =head1 ACKNOWLEDGEMENTS
 
 L<AirG Inc.|http://corp.airg.com>, my employer, and the first user of this distribution.
@@ -474,14 +638,21 @@ L<LWP::UserAgent>
 
 L<PSGI>, L<HTTP::Message::PSGI>
 
-=head1 COPYRIGHT
+=head1 AUTHOR
 
-This software is copyright (c) 2012 by Karen Etheridge, <ether@cpan.org>.
+Karen Etheridge <ether@cpan.org>
 
-=head1 LICENSE
+=head1 COPYRIGHT AND LICENSE
+
+This software is copyright (c) 2012 by Karen Etheridge.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
 
 =cut
 
+
+__END__
+
+
+1;
